/* This version of grumpy uses the device filesystem */

#include "linux/config.h"
#ifdef CONFIG_SMP
	#define __SMP__
#endif
#include "linux/kernel.h"
#include "linux/module.h"
#include "linux/fs.h"
#include "asm/uaccess.h"  /* for put_user */
#include "linux/devfs_fs_kernel.h"  /* device filesystem calls */
	
int init_module(void);
void cleanup_module(void);
static int device_open(struct inode *, struct file *);
static int device_release(struct inode *, struct file *);
static ssize_t device_read(struct file *, char *, size_t, loff_t *);
static ssize_t device_write(struct file *, const char *, size_t, loff_t *);
	
#define SUCCESS 0
#define DEVICE_NAME "chardev" /* Dev name as it appears in /proc/devices   */
#define BUF_LEN 80            /* Max length of the message from the device */
	
	
/* Global variables are declared as static, so are global within the file. */
	
static int Major;         /* Major number assigned to our device driver */
static int Device_Open=0; /* Is device open?  Prevent multiple access to file */
static char msg[BUF_LEN]; /* The msg the device will give when asked */
static char *msg_Ptr;
static struct file_operations fops = {
  .read = device_read, 
  .write = device_write,
  .open = device_open,
  .release = device_release
};
	
	

int init_module(void)
{
	devfs_handle_t grumpy_devfs_dir;
	if (! grumpy_devfs_dir) /* Houston control, we have a problem... */
		return -EBUSY;

	devfs_register(grumpy_devfs_dir, "grumpy", DEVFS_FL_DEFAULT, 0, 0,
		S_IFCHR | S_IRUGO | S_IWUGO, &fops, 1);
		
   Major = register_chrdev(0, DEVICE_NAME, &fops);
   if (Major < 0) {
     printk ("Registering the character device failed with %d\n", Major);
     return Major;
   }
	
   printk("<1>Assigned major number %d.  To talk to the driver,\n", Major);
   printk("<1>create a dev file with 'mknod /dev/grumpy c %d 0'.\n", Major);
   printk("<1>Try various minor numbers.  Try to cat and echo to the device\n");
	printk("file.  Remove the device file and module when done.\n");
	
   return 0;
}
	
	
void cleanup_module(void)
{
   int ret = unregister_chrdev(Major, DEVICE_NAME);
   if (ret < 0) printk("Oh, shit!  I can't unregister grumpy: %d\n", ret);
}  
	
	

/* Called when a process tries to open the device file, like
 * "cat /dev/mycharfile"
 */
static int device_open(struct inode *inode, struct file *file)
{
   static int count = 0;

   if (Device_Open) return -EBUSY;

   Device_Open++;
   snprintf(msg, BUF_LEN, "I already told you %d damned times.  Hello world!!\n",count++);

   msg_Ptr = msg;
   MOD_INC_USE_COUNT;
	
  return SUCCESS;
}
	
	
/* Called when a process closes the device file.
 */
static int device_release(struct inode *inode, struct file *file)
{
	Device_Open --;     /* We're now ready for our next caller */

	/* Decrement the usage count, or else once you opened the file, you'll
		 never get get rid of the module.  Reboot city! */
	MOD_DEC_USE_COUNT;

	return 0;
}


/* Called when a process, which already opened the dev file, attempts to
	read from it.
*/
static ssize_t device_read(struct file *filp,
	char *buffer,    /* The buffer to fill with data */
	size_t length,   /* The length of the buffer     */
	loff_t *offset)  /* Our offset in the file       */
{
	int bytes_read = 0; /* Number of bytes written to the buffer */

	/* If we're at the end of the message, return 0 signifying end of file */
	if (*msg_Ptr == 0) return 0;

	while (length && *msg_Ptr)  {
		/* Put data into the buffer. However, msg is in user data segment, not the
		 *	not the kernel segment; assignment won't work.  put_user copies data
		 *	from kernel data segment to user data segment.
		 */
		put_user(*(msg_Ptr++), buffer++);

		length--;
		bytes_read++;
	}

	return bytes_read; /* Return the # of bytes put into the buffer */
}


/*  Called when a process writes to dev file:  echo "hi" > /dev/hello
 */
static ssize_t device_write(struct file *filp, const char *buff, size_t len,
	 loff_t *off)
{
	printk ("<1>Write all over me?  I'll show you.  Take this!\n");
	printk ("<1>Kernel Oops detected.  Kernel dumps core.\n");
	return -EINVAL;
}
